/* -*- mode: c++; c-basic-offset: 2 -*- */

#include "jengine.h"
#include "jplus.h"

#include <set>
#include <map>
#include <stack>
#include <utility>
#include <cassert>
#include <iostream>
#include <iomanip>

using namespace std;

bool TRACE=false; /**< enable/disable execution tracing to stderr */
bool TRACETOKEN=false; /**< enable/disable tokenizer tracing to stderr */
bool TRACESSA=false; /**< enable/disable SSA tracing to stderr */

////////////////////////////////////////////////////////////////////////////////
// === J sentence tokenizer ----------------------------------------------------
// character classes
static const int CX=0;  /* other */
static const int CS=1;  /* space or tab */
static const int CA=2;  /* letter */
static const int CN=3;  /* N (capital N) */
static const int CB=4;  /* B (capital B) */
static const int C9=5;  /* digit or sign (underscore) */
static const int CD=6;  /* dot */
static const int CC=7;  /* colon */
static const int CQ=8;  /* quote */

// character classification
const char wtype[256]={
  0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,CS,CS,0 ,0 ,0 ,0 ,0 , /* 0                  */
  0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 , /* 1                  */
  CS,0 ,0 ,0 ,0 ,0 ,0 ,CQ,0 ,0 ,0 ,0 ,0 ,0 ,CD,0 , /* 2  !"#$%&'()*+,-./ */
  C9,C9,C9,C9,C9,C9,C9,C9,C9,C9,CC,0 ,0 ,0 ,0 ,0 , /* 3 0123456789:;<=>? */
  0 ,CA,CB,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CN,CA, /* 4 @ABCDEFGHIJKLMNO */
  CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,0 ,0 ,0 ,0 ,C9, /* 5 PQRSTUVWXYZ[\]^_ */
  0 ,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA, /* 6 `abcdefghijklmno */
  CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,CA,0 ,0 ,0 ,0 ,0 , /* 7 pqrstuvwxyz{|}~  */
};
/* 0, 1  2  3  4  5  6  7  8  9  a  b  c  d  e  f   */

// states
static const int SS=0;  /* space */
static const int SX=1;  /* other */
static const int SA=2;  /* alphanumeric */
static const int SN=3;  /* N */
static const int SB=4;  /* NB */
static const int SD=5;  /* NB. */
static const int S9=6;  /* numeric */
static const int SQ=7;  /* quote */
static const int SO=8;  /* even quotes */
static const int SZ=9;  /* trailing comment */
  
// actions
static const int EI =1;  /* emit(b,i-1) */
static const int EN =2;  /* b=i */
static const int EV =3;  /* emit vector*/
  
typedef struct {char next,action;} ST;

// state transition table
static const ST state[10][9]={  
/*SS*/{{SX,EN},{SS,0 },{SA,EN},{SN,EN},{SA,EN},{S9,EN},{SX,EN},{SX,EN},{SQ,EN}},
/*SX*/{{SX,EI},{SS,EI},{SA,EI},{SN,EI},{SA,EI},{S9,EI},{SX,0 },{SX,0 },{SQ,EI}},
/*SA*/{{SX,EI},{SS,EI},{SA,0 },{SA,0 },{SA,0 },{SA,0 },{SX,0 },{SX,0 },{SQ,EI}},
/*SN*/{{SX,EI},{SS,EI},{SA,0 },{SA,0 },{SB,0 },{SA,0 },{SX,0 },{SX,0 },{SQ,EI}},
/*SB*/{{SX,EI},{SS,EI},{SA,0 },{SA,0 },{SA,0 },{SA,0 },{SD,0 },{SX,0 },{SQ,EI}},
/*SD*/{{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SX,0 },{SX,0 },{SZ,0 }},
/*S9*/{{SX,EI},{SS,EV},{S9,0 },{S9,0 },{S9,0 },{S9,0 },{S9,0 },{SX,0 },{SQ,EI}},
/*SQ*/{{SQ,0 },{SQ,0 },{SQ,0 },{SQ,0 },{SQ,0 },{SQ,0 },{SQ,0 },{SQ,0 },{SO,0 }},
/*SO*/{{SX,EI},{SS,EI},{SA,EI},{SN,EI},{SA,EI},{S9,EI},{SX,EI},{SX,EI},{SQ,0 }},
/*SZ*/{{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 },{SZ,0 }},
};
/*       CX      CS      CA      CN      CB      C9      CD      CC      CQ   */

// first letter of OP classification
const int L1=(1<<7); // can have length 1
const int L2D=(1<<6); // length 2 and dot at the end decide it as operator
const int L2C=(1<<5); // length 2 and colon at the end decide it as operator
const int S3= 1|L1|L2D|L2C; // alone, followed by '.' or ':'
const int S2= 2|L1|L2C; // alone, or followed by ':'
const int Q2= 3|L1|L2D; // alone, or followed by '.'
const int S4= 4|L1|L2D|L2C; // alone, followed by '.', ':' or '.:'
const int B4= 5|L1|L2D|L2C; // alone, followed by '.', ':' or '::'
const int A2= 6|L2D|L2C; // followed by '.' or ':', can't be alone
const int A1= 7|L2D; // followed by '.'
const int B1= 8|L2C; // followed by ':'
const int A3= 9|L2D|L2C; // followed by '.',':' or '..'
const int S0=10|L1; // alone only

const char otype[256]={
  0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 , /* 0                  */
  0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 , /* 1                  */
  0 ,S3,S3,S3,S3,S3,S4,0 ,0 ,0 ,S3,S3,S3,S3,S3,S3, /* 2  !"#$%&'()*+,-./ */
  0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 ,S3,S3,S3,S3,S3,Q2, /* 3 0123456789:;<=>? */
  S3,A1,0 ,A1,A2,A1,0 ,0 ,A1,A1,0 ,0 ,A2,A1,0 ,0 , /* 4 @ABCDEFGHIJKLMNO */
  0 ,0 ,0 ,B1,A1,0 ,0 ,0 ,0 ,0 ,0 ,S2,S3,S0,S3,S3, /* 5 PQRSTUVWXYZ[\]^_ */
  S2,A2,A1,0 ,A1,A1,A1,0 ,0 ,A2,A1,0 ,0 ,0 ,0 ,A1, /* 6 `abcdefghijklmno */
  A3,B1,A1,B1,A2,B1,0 ,0 ,B1,0 ,0 ,B4,S3,S3,S3,0 , /* 7 pqrstuvwxyz{|}~  */
};
/* 0, 1  2  3  4  5  6  7  8  9  a  b  c  d  e  f   */

// types of tokens
const int TOK_SDELIM=(1<<12); // sentence delimiter
const int TOK_LITERAL=(1<<11);

const int TOK_ERROR=-1; // spelling error
const int TOK_EOLN=1 | TOK_SDELIM;
const int TOK_OP =2;
const int TOK_NUM =3 | TOK_LITERAL;
const int TOK_STR =4 | TOK_LITERAL;
const int TOK_ID =5;
const int TOK_LBRACE =7;
const int TOK_RBRACE =8;
const int TOK_KEYWORD =9 | TOK_SDELIM;
const int TOK_GOTOORLABEL =10 | TOK_SDELIM;

class Tokenizer {
public:
  int s;
  char ch;
  istream &in;
  
  static set<string> keywords;
  
  Tokenizer(istream &in_):in(in_),s(SS),ch(' ') {
    if(keywords.size()==0) { // initialize keywords list
      keywords.insert("assert."); keywords.insert("break.");
      keywords.insert("case."); keywords.insert("catch.");
      keywords.insert("catchd."); keywords.insert("catcht.");
      keywords.insert("continue."); keywords.insert("do.");
      keywords.insert("else."); keywords.insert("elseif.");
      keywords.insert("end."); keywords.insert("fcase.");
      keywords.insert("for."); keywords.insert("if.");
      keywords.insert("select."); keywords.insert("throw.");
      keywords.insert("try."); keywords.insert("while.");
      keywords.insert("whilst.");
    };
  }
  
  bool good() {
    return in.good();
  }

  static char tokenType(int type) {
    if (type<0) return 'E';
    if (type&TOK_SDELIM) return ';';
    if (type&TOK_LITERAL) return '0';
    if (type==TOK_ID) return 'a';
    if (type==TOK_OP) return '+';
    if (type==TOK_LBRACE) return '(';
    if (type==TOK_RBRACE) return ')';
    return '?';
  };

  static bool isValidID(string tok) {
    // check just in case that ID's start with letter and contain only letters
    // or digits
    char c;
    c=tok[0]; if ((c<0)||(c>255)) c==0;
    bool LOD=(wtype[c]==CA)||(wtype[c]==CN)||(wtype[c]==CB);
    for (int i=1;(i<tok.size())&&LOD;i++) {
      c=tok[i]; if ((c<0)||(c>255)) c==0;
      LOD = (wtype[c]==CA)||(wtype[c]==CB)||(wtype[c]==CN)|| (wtype[c]==C9);
    };
    return LOD;
  };

  // return token type or TOK_ERROR in case of spelling error
  static int classifyToken(string tok) {
    // recognize operators EOLN and braces
    char c,ED,EC;
    switch (tok.size()) {
    case 0:
      return TOK_EOLN;
    case 1:
      // can be operator
      c=tok[0]; 
      if (c=='(') return TOK_LBRACE;
      if (c==')') return TOK_RBRACE;
      if ((c<0)||(c>255)) c==0;
      if (otype[c]&L1) return TOK_OP;
      break; // otherwise not an operator
    case 2:
      ED=(tok[1]=='.'?L2D:0);
      EC=(tok[1]==':'?L2C:0);
      c=tok[0]; if ((c<0)||(c>255)) c==0;
      if ((otype[c]&ED)||(otype[c]&EC)) return TOK_OP;
      break; // otherwise not an operator
    case 3:
      // here we have three remaining cases: S4, B4 and A3
      c=tok[0]; if ((c<0)||(c>255)) c==0;
      if ((otype[c]==A3)&&(tok[1]=='.')&&(tok[2]=='.')) return TOK_OP;
      if (((tok[1]=='.')||(tok[1]==':'))&&
	  (((otype[c]==S4)||(otype[c]==B4))&&(tok[2]==':'))) return TOK_OP;
      break; // otherwise not an operator
    default:
      break;
    };
    // not operator, but can be keyword, label, literal or ID
  
    // all keywords and labels start with letter and end with dot
    //  assert. break. case. catch. catchd. catcht. continue. do.
    //  else. elseif. end. fcase. for. goto_name. label_name. if.
    //  select. throw. try. while. whilst.
    if (tok.size()>=3) { // the shortest ones are "do.", "if."
      c=tok[0]; if ((c<0)||(c>255)) c==0;
      if ((wtype[c]==CA) &&(tok[tok.size()-1]=='.')) {
	if (keywords.find(tok) != keywords.end()) return TOK_KEYWORD;
	// it still can be goto or label
	if ((tok.substr(0,5)=="goto_")&&
	    (isValidID(tok.substr(5,tok.size()-5-1)))) return TOK_GOTOORLABEL;
	if ((tok.substr(0,6)=="label_")&&
	    (isValidID(tok.substr(6,tok.size()-6-1)))) return TOK_GOTOORLABEL;
	return TOK_ERROR;
      };
    };
  
    // now it's either literal or ID.
    // literals start with number or single quote (string literals)
    if (wtype[c]==C9) return TOK_NUM;
    if (wtype[c]==CQ) return TOK_STR;
    
    // otherwise it's an ID
    if (isValidID(tok)) return TOK_ID;

    // or spelling error
    return TOK_ERROR;
  };

  static char wtypef(const char c) {
    if ((c<0)||(c>255)) return CX; else return wtype[c];
  };
  
  // returns the next token or empty string if there is EOL or EOF
  string next() {
    if (!in.good()||(ch=='\n')) {
      ch=' ';
      return "";
    };
    string buf(1,ch);
    int b=0;
    ch=in.get();
    while (true) {
      int cht=wtypef(ch);
      if ((s==SZ) && (ch=='\n')) { buf.clear(); b=0; ch=' '; s=SS; break; };
      ST p=state[s][cht]; int e=p.action;
      if (!in.good()) e=EI;
      if (e==EV) {
	// don't emit just yet, skip space and look if the next element
	// is not a number
	while (in.good()&&(ch!='\n')&&((cht=wtypef(ch))==CS)) ch=in.get();
	e=EI; 
	if (in.good()&&(ch!='\n')) {
	  if (cht==C9) { 
	    buf.append(1,' '); p=state[s][cht]; e=p.action;
	  } else { 
	    in.putback(ch); ch=' '; p=state[s][CS];
	  };
	};
      }
      s=p.next;
      if (e==EI) { b=0; break; };
      if (e) {buf.clear(); b=0;};
      buf.append(1,ch);
      ch=in.get();
    };
    if (wtypef(buf[b])==CS) b++;
    if (buf[b]=='\n') b++;
    return buf.substr(b,buf.size()-b);
  }
};
set<string> Tokenizer::keywords;
// === J sentence tokenizer ----------------------------------------------------
////////////////////////////////////////////////////////////////////////////////

// a subscripted variable
class Var: public pair<string,int> {
  friend ostream& operator<<(ostream& output, const Var& v);
public:
  Var(): pair<string,int>("",-1) {}
  Var(string n): pair<string,int>(n,-1) {}
  Var(string n, int s): pair<string,int>(n,s) {}
  const string& getName() const {return first; }
  int getS() const {return second; }
  void setS(int s) {second=s; }
  string toString() const {
    string r(first);
    if (second>0) {
      r+=(static_cast<ostringstream*>( &(ostringstream() << second) )->str());
      return r;
    };
  };
};
ostream& operator<<(ostream& output, const Var& v) {
  output << v.first;
  if (v.second>0) output << '_' << v.second;
  return output;
}

// a single sentence
class Sentence {
public:
  //  string J; // J sentence, defining this vertex
  vector<string> tokens; // J sentence, tokenized
  vector<Var> vars;      // variable references in J sentence
  vector<int> ttype;     // token types
  vector<int> sentE;     // ends of subsentences, separated by brackets
  vector<int> sentS;     // beginnings of subsentences, separated by brackets
  map<Var,set<Var> > outDeps; // input variables, influencing given output
  map<Var,set<Var> > inDeps; // output variables, influenced by given input
  
  // local numbering of the variables, just in this sentence
  map<string,int> local_numbering;

  // ancestors in flow control graph
  vector<int> ancestors;

  Sentence() {
  };

  // read sentence
  string read(string tok, Tokenizer &lex) {
    // first we read all tokens of the sentence
    tokens.clear();
    int tt;
    if (TRACETOKEN) cerr << endl << "BEGINNING of sentence" << endl;
    if (TRACETOKEN) cerr <<'\''<<tok<<'\''
			 <<lex.tokenType(lex.classifyToken(tok));
    while(((tt=lex.classifyToken(tok))&TOK_SDELIM)==0) {
      if (TRACETOKEN) cerr <<'\''<<tok<<'\''<<lex.tokenType(tt);
      if ((tt&TOK_SDELIM)) cerr << '!';
      if (TRACETOKEN) cerr<<"->";
      tokens.push_back(tok);
      ttype.push_back(tt);
      tok=lex.next();
      if (TRACETOKEN) 
	cerr << '\'' << tok << '\'' <<lex.tokenType(lex.classifyToken(tok)) 
	     << ' ';
    };
    if (TRACETOKEN) cerr << endl << "END of sentence" << endl;
    vars.clear(); vars.resize(tokens.size());

    // next we calculate the end of each sub-sentence, separated by brackets
    sentE.clear(); sentE.resize(tokens.size(),-1);
    sentS.clear(); sentS.resize(tokens.size(),0);
    stack<int> beginning;
    beginning.push(0);
    for (int i=0;i<=tokens.size();i++) {
      const string &tok=(i<tokens.size())?tokens[i]:")";
      if (tok=="(") beginning.push(i);
      if (tok==")") {
	for(int j=beginning.top();j<i;j++)
	  if (sentE[j]<0) sentE[j]=i;
	beginning.pop();
      };
      if (i<tokens.size()) sentS[i]=beginning.top();
    };
    assert(beginning.size()==0);
    // TODO: behave more graciously in case of mismatched brackets

    return tok;
  };

  void print() {
    for(vector<string>::iterator it=tokens.begin();it!=tokens.end();it++)
      cerr << *it << " ";
    cerr << "  [out| ";
    for(map<Var,set<Var> >::iterator it=outDeps.begin();it!=outDeps.end();it++){
      if (it!=outDeps.begin()) cerr << ", ";
      cerr << (*it).first << '{';
      const set<Var> &deps=(*it).second; 
      for(set<Var>::iterator jt=deps.begin();jt!=deps.end();jt++) {
	if (jt!=deps.begin()) cerr << ',';
	cerr << (*jt);
      };
      cerr << '}'; 
    };
    cerr << "  [in|";
    for(map<Var,set<Var> >::iterator it=inDeps.begin();it!=inDeps.end();it++){
      if (it!=inDeps.begin()) cerr << ", ";
      cerr << (*it).first << '{';
      const set<Var> &deps=(*it).second; 
      for(set<Var>::iterator jt=deps.begin();jt!=deps.end();jt++) {
	if (jt!=deps.begin()) cerr << ',';
	cerr << (*jt);
      };
      cerr << '}'; 
    };
    cerr << endl;
  };

  void add_ancestor(int n) {
    ancestors.push_back(n);
  };

  int get_index(string vn, vector<Sentence> &sentences) {
    map<string,int>::iterator it;
    it=local_numbering.find(vn);
    if (it==local_numbering.end()) {
      if (ancestors.size()>0) {
	// check ancestors
	int n=sentences[ancestors[0]].get_index(vn, sentences);
	for (int i=1; i<ancestors.size(); i++) {
	  int n1=sentences[ancestors[i]].get_index(vn, sentences);
	  assert(n1=n);
	};
	return n;
      } else return -1;
    }
    return (*it).second;
  };

  void assign_subscripts_int(int &i, set<Var> &cdeps, 
			     vector<Sentence> &sentences) {
    // first, if we have several ancestors, we need to add a virtual phi nodes
    // currently we may have at most one ancestor, no phi nodes will be inserted
    assert(ancestors.size()<=1);
    
    int l=sentS[i]; // current label
    for(;(i>l);i--) {
      assert(sentS[i]==l);
      if (sentS[i-1]>sentS[i]) {
	i--;
	set<Var> ncdeps;
	assign_subscripts_int(i,ncdeps,sentences); // dive
	cdeps.insert(ncdeps.begin(),ncdeps.end()); // merge deps
      };
      if (tokens[i]=="=:") {
	if (ttype[i-1]==TOK_ID) { // we have an output variable
	  const string &vn=tokens[i-1];
	  Var v(vn,get_index(vn, sentences)+1); // increment subscript
	  if (TRACESSA)
	    cerr << "SSA newvar " << v << endl;
	  outDeps.insert(pair<Var,set<Var> >(v,cdeps));
	  cdeps.clear();
	  local_numbering.erase(v.getName());
	  local_numbering.insert(v);
	  if (i==1) {
	    vars[i-1]=v;
	    tokens[i-1]=v.toString(); // patch the number into
	  };
	} else {
	  cerr << "assignment must be preceded by an explicit noun name.";
	  assert(false);
	};
      };
      if (tokens[i]=="=.") {
	cerr << "local assignments are not currently allowed.";
	assert(false);
      };    
      if (ttype[i]==TOK_ID) {
	// we have an input variable
	const string &vn=tokens[i];
	Var v(vn,get_index(vn, sentences));
	if (v.getS()<0) {
	  cerr << "Variable \"" << vn << "\" is not defined.";
	  assert(false);
	};
	cdeps.insert(v);
	vars[i]=v;
	tokens[i]=v.toString(); // patch the number into
      };
    };
  }; 
  
  bool assign_subscripts(vector<Sentence> &sentences) {
    // if there are several ancestors, we need to add a virtual phi nodes
    // currently we may have at most one ancestor, no phi nodes will be inserted
    assert(ancestors.size()<=1);
    set<Var> cdeps; // currently accumulated input deps
    int i=tokens.size()-1;
    if (i>=0) {
      assign_subscripts_int(i, cdeps, sentences);
      assert(i==0);
    };
    // we also compute here the reverse deps (what output variables change
    // if a particular input variable is modified). We do it just by reversing
    // the outDeps map.
    for(map<Var,set<Var> >::iterator it=outDeps.begin();it!=outDeps.end();it++){
      const Var &o=(*it).first;
      const set<Var> &deps=(*it).second; 
      for(set<Var>::iterator jt=deps.begin();jt!=deps.end();jt++) {
	const Var &i=(*jt);
	// here we have the dependence of an output variable "o" on input "i"
	map<Var,set<Var> >::iterator r=inDeps.find(i);
	if (r==inDeps.end()) { // insert key
	  set<Var> oset;
	  oset.insert(o);
	  inDeps.insert(pair<Var,set<Var> >(i,oset));
	} else  (*r).second.insert(o); // add to existing
      };
    };
    

    return true; // a single sentence can always be subscripted in one go
  };

  // set of variables, defined by this sentence
  set<Var> outVars() {
    set<Var> res;
    for(map<Var,set<Var> >::iterator it=outDeps.begin();it!=outDeps.end();it++)
      res.insert((*it).first);
    return res;
  };

  // this sentence defines the variable v
  bool defines(const Var &v) {
    map<Var,set<Var> >::iterator r=outDeps.find(v);
    return r!=outDeps.end();
  };

  // this sentence defines the variable v
  bool influencedBy(const Var &v) {
    map<Var,set<Var> >::iterator r=inDeps.find(v);
    return r!=inDeps.end();
  };

  // this sentence defines any of the set v
  bool definesAny(const set <Var> &v) {
    bool res=false;
    for (set<Var>::iterator it=v.begin(); (it!=v.end())&&(!res);it++) {
      map<Var,set<Var> >::iterator r=outDeps.find(*it);
      res= (r!=outDeps.end());
    };
    return res;
  };

  // this sentence depends on the variable v
  bool dependsOn(const Var &v) {
    map<Var,set<Var> >::iterator r=inDeps.find(v);
    return r!=inDeps.end();
  };

  // set of variables, on which "v" depends
  set<Var> dependenciesOf(const Var &v) {
    map<Var,set<Var> >::iterator r=outDeps.find(v);
    assert(r!=outDeps.end());
    return (*r).second;
  };

  // set of variables, which depend on "v" in this sentence
  set<Var> dependentsOf(const Var &v) {
    map<Var,set<Var> >::iterator r=inDeps.find(v);
    assert(r!=inDeps.end());
    return (*r).second;
  };

  void emit(int &i, bool emitting, vector<string> &program,set<Var> &lvars, 
	    set<Var> &dvars) {
    string J="";
    int l=sentE[i-1]; // current label
    for(;(i<l);i++) {
      assert(sentE[i-1]==l);
      if (sentE[i-1]>sentE[i]) {
	if (emitting) J+=tokens[i]+" ";
        i++;
	emit(i,emitting,program,lvars,dvars); // dive
	if (emitting) { // embed into the current sentence
	  J+=program.back();
	  program.pop_back();
	};
      };
      if (tokens[i]=="=:") {
	assert(ttype[i-1]==TOK_ID); // we have an output variable
	if ((!emitting)&&(lvars.count(vars[i-1])>0)) J=tokens[i-1]+ " ";
	emitting=(lvars.count(vars[i-1])>0);
      };
      if (emitting) J+=tokens[i]+" ";
    };
    if (J.size()>0) program.push_back(J);
  };
  
  void emit(vector<string> &program, set<Var> &lvars, set<Var> &dvars) {
    if (tokens.size()==0) return;
    int i=1;
    emit(i, false, program, lvars, dvars);
    assert(i==tokens.size()); // check that we came to an end
  };
 
};

// block: a sequence of sentences
class BasicBlock {
public:
  
  vector<Sentence> sentences;
  set<Var> allVars;        // all variables
  map<Var,int> definedIn;  // sentence, where a given var is defined
  set<Var> dirtyVars;     // "dirty" variables, need to be recalculated

  BasicBlock(const jengine &eng) {
    Sentence lib; // dummy sentence, defining all builtin functions
    set<string> builtins=eng.getBuiltins();
    set<string>::iterator it;
    for(it=builtins.begin(); it!=builtins.end(); it++) {
      Var v(*it,0);
      set<Var> cdeps; // empty set
      lib.outDeps.insert(pair<Var,set<Var> >(v,cdeps));
      lib.local_numbering.insert(v);
    };
    sentences.push_back(lib);
  }

  // read block
  string read(string tok, Tokenizer &lex) {
    do {
      if (tok.size()>0) {
	Sentence nv;
	tok=nv.read(tok, lex);
	if (sentences.size()>0) nv.add_ancestor(sentences.size()-1);
	sentences.push_back(nv);
      };
      tok=lex.next();
    } while(lex.good());
    return tok; // which is, obviously, empty string
  }

  bool assign_subscripts() {
    for (int i=0; i<sentences.size(); i++) {
      sentences[i].assign_subscripts(sentences);
      if (TRACE) sentences[i].print();
      set<Var> outs=sentences[i].outVars();
      allVars.insert(outs.begin(),outs.end());
      for (set<Var>::iterator it=outs.begin(); it!=outs.end(); it++) {
	assert(definedIn.count(*it)==0); // variables are defined only once
	definedIn.insert(pair<Var,int>(*it,i));
      };
    };
    // at the beginning all "out" variables are not current
    dirtyVars.insert(allVars.begin(),allVars.end());
  }

  typedef void (*links)(void *inst, const Var &v, set<Var> &more);

  void closure(set<Var> &vars, links f) {
    set<Var> needed;
    needed.insert(vars.begin(),vars.end());
    vars.clear();
    while (!needed.empty()) {
      const Var &v=*(needed.begin());
      if (vars.insert(v).second)
	f(this,v,needed);
      needed.erase(v);
    };
  }

  static void linkDeps(void *inst, const Var &v, set<Var> &more) {
    BasicBlock *b=(BasicBlock *)inst;
    int defs=0;
    for (int i=0; i<(b->sentences.size());i++) 
      if (b->sentences[i].defines(v)) {
	set<Var> deps= b->sentences[i].dependenciesOf(v);
	more.insert(deps.begin(), deps.end());
	defs++;
      };
    assert(defs==1);
  }

  static void linkModifys(void *inst, const Var &v, set<Var> &more) {
    BasicBlock *b=(BasicBlock *)inst;
    for (int i=0; i<(b->sentences.size());i++) 
      if (b->sentences[i].influencedBy(v)) {
	set<Var> deps= b->sentences[i].dependentsOf(v);
	more.insert(deps.begin(), deps.end());
      };
  }

  const static int MAX=10000;

  // constructs and returns the var with specified subscript
  // the subscript can be BasicBlock::MAX to specify that the last
  // version of the variable (as defined in this block) is required.
  Var getVar(string name, int subscript=BasicBlock::MAX) {
    int idx=sentences.back().get_index(name,sentences);
    if (subscript!=MAX) idx=subscript;
    Var v(name,idx);
    return v;
  }

  // marks the variable with the given name (and smallest index) as
  // "changed" as well as all the variables, which depend on it.
  void propagateChanged(const set<Var> &vars) {
    // calculate live variables and sentences
    dirtyVars.insert(vars.begin(),vars.end());
    closure(dirtyVars,linkModifys);
    for(set<Var>::iterator it=vars.begin(); it!=vars.end(); it++)
      dirtyVars.erase((*it)); // note that we have already set vars externally
    
    if (TRACE) {
      cerr << "Changing ";
      for(set<Var>::iterator it=vars.begin(); it!=vars.end(); it++)
	cerr << (*it) << ' ';
      cerr << ".  Dirty: ";
      for(set<Var>::iterator it=dirtyVars.begin(); it!=dirtyVars.end(); it++)
	cerr<<(*it)<<" ";
      cerr << endl;
    };
  }

  // pulls sentences (in form of strings) to compute the given variable
  void pullSentences(const set<Var> &vars,  vector<string> &res) {
    // calculate live variables and sentences
    set<Var> liveVars;
    liveVars.insert(vars.begin(),vars.end());
    closure(liveVars,linkDeps);
    
    // ensure that only "dirty" ones are recalculated
    set<Var> toCompute;
    for(set<Var>::iterator it=liveVars.begin(); it!=liveVars.end(); it++)
      if (dirtyVars.count(*it)>0) 
	toCompute.insert(*it);
    
    // determine live statements
    // TODO: use "definedIn" for speed up
    vector<bool> liveStmts(sentences.size(),false);
    for(int i=0;i<sentences.size();i++)
      liveStmts[i]=sentences[i].definesAny(toCompute);
    
    // generate the program
    for(int i=0;i<sentences.size();i++)
      if (liveStmts[i]) sentences[i].emit(res,toCompute,dirtyVars);
    
    // retire "dirty" variables
    for(set<Var>::iterator it=dirtyVars.begin(); it!=dirtyVars.end(); ) 
      if (toCompute.count(*it)>0) 
	dirtyVars.erase(it++);
      else
	it++;

    if (TRACE) {
      cerr << "--- To compute ";
      for(set<Var>::iterator it=vars.begin(); it!=vars.end(); it++)
	cerr << (*it) << ' ';
      cerr << "we need :" << endl;
      for (int i=0; i<res.size();i++)
	cerr << res[i] << endl;
      cerr << "------ " << res.size() << " lines.  Dirty: ";
      for(set<Var>::iterator it=dirtyVars.begin(); it!=dirtyVars.end(); it++)
	cerr<<(*it)<<" ";
      cerr << endl;
    }
  }

};

jplus::jplus() {
  program=NULL;
};

bool jplus::init(std::istream &script) {
  libInit();
  program=new BasicBlock(*this);
  //   parseStream(script);
  Tokenizer lex(script);
  string last_tok;
  bool ok=((last_tok=program->read(lex.next(), lex))).size()==0; // last token is EOF
  if (TRACETOKEN)
    cerr << "last token = "<< '"'<< last_tok << '"' << endl;
  if (ok) program->assign_subscripts(); // convert to SSA form
  return ok;
};

jplus::~jplus() {
  if (program) delete program;
};

void jplus::libInit() {};

//bool jplus::parseStream(std::istream &script) {
//  Tokenizer lex(script);
//  return (program->read(lex.next(), lex)).size()==0; // last token is EOF
//}

bool jplus::set(std::string name, jarray data) {
  std::set<Var> vars;
  Var v=program->getVar(name,0); // updates are done to the first version
  vars.insert(v);
  program->propagateChanged(vars);
  return jengine::set(v.toString(),data);
}

jarray jplus::get(std::string name) {
  std::set<Var> vars;
  Var v=program->getVar(name,BasicBlock::MAX);// we always get the last version
  vars.insert(v);
  vector <string> res;
  program->pullSentences(vars,res);
  for(int i=0; i<res.size(); i++)
    doJ(res[i]);
  return jengine::get(v.toString());
}

/** pull minimal set of J sentences for computing a given set of variables, 
 *  assuming everything is "dirty".
 */
std::vector<std::string> jplus::getProgram(std::set<std::string> varnames) {
  // compute the last versions of specified variables
  std::set<Var> vars;
  for(std::set<std::string>::iterator it=varnames.begin(); 
      it!=varnames.end(); it++) {
    Var v=program->getVar((*it),BasicBlock::MAX);
    vars.insert(v);
  };
  // make all vars dirty temporarily
  std::set<Var> oldDirtyVars=program->dirtyVars;
  program->dirtyVars.insert(program->allVars.begin(),program->allVars.end());
  // pull the sentences
  vector <string> res;
  program->pullSentences(vars,res);
  // restore dirty variables
  program->dirtyVars=oldDirtyVars; 
  return res;
};

/** pull all J in jplus script. This can be done in a much more direct
    way, without using all the jplus machinery. The more complex way
    is chosen to test jplus and fail early if there is a bug.
 */
std::vector<std::string> jplus::getProgram() {
  // compute all the var names, encountered
  std::set<std::string> varnames;
  std::set<Var> &vars=program->allVars;
  for(std::set<Var>::iterator it=vars.begin();it!=vars.end(); it++)
    varnames.insert((*it).getName());
  return getProgram(varnames);
};
